# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import unittest
from unittest.mock import patch, MagicMock
from io import StringIO
import json
from . import core
import datetime
import requests

class MockResponse(object):
    def __init__(self, body, status_code):
        self.body = body
        self.status_code = status_code

    def text(self):
        return self.body

    def json(self):
        return json.loads(self.body)

class SchedyDBTest(unittest.TestCase):
    def setUp(self):
        config = StringIO('{"root": "schedy://schedy.io/", "email": "test@schedy.io", "token": "token"}')
        self.db = core.SchedyDB(config)

    def test_config(self):
        self.assertEqual(self.db.root, 'schedy://schedy.io/')
        self.assertEqual(self.db.email, 'test@schedy.io')
        self.assertEqual(self.db.api_token, 'token')

    @patch('requests.Session')
    def test_authenticate(self, sess_mock):
        # Test can authenticate
        exp_date = datetime.datetime.now() + datetime.timedelta(days=1)
        resp_mock = sess_mock.return_value.request.return_value
        resp_mock.status_code = 200
        resp_mock.json.return_value = {'token': 'jwt.token.valid', 'expiresAt': exp_date.timestamp()}
        self.db._authenticate()
        self.assertEqual(self.db._jwt_token.token_string, 'jwt.token.valid')
        self.assertEqual(self.db._jwt_token.expires_at, exp_date)
        # Test does not request token second time
        resp_mock.json.return_value = {}
        resp = self.db._authenticated_request('GET', '/')
        # Test session not recreated
        sess_mock.assert_called_once()
        # Test not reauthenticated
        sess_mock.return_value.request.assert_('GET')




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

import requests

class SchedyError(Exception):
    '''
    Base class for all Schedy exceptions.
    '''
    pass

class HTTPError(SchedyError):
    '''
    Base class for exceptions caused by a transaction with the service.
    '''
    def __init__(self, body, code, *args):
        '''
        Args:
            body (str): Error message.
            code (int or None): HTTP status code.
            args (list): Other arguments passed to :py:exc:`Exception`.
        '''
        self.code = code
        self.body = body
        # Truncate long bodies
        msg_body = body
        msg_body = msg_body.replace('\r', '')
        msg_body = msg_body.rstrip('\n')
        if len(msg_body) == 0:
            msg_body = '<No server message>'
        elif len(msg_body) > 150:
            msg_body = msg_body[:147] + '...'
        msg_body = msg_body.replace('\n', '\n> ')
        message = 'HTTP Error {}:\n> {}'.format(code, msg_body)
        super().__init__(message, *args)

class ClientError(HTTPError):
    '''
    Exception caused by the client side.
    '''
    pass

class ClientRequestError(ClientError):
    '''
    Exception caused by the content of the request.
    '''
    pass

class AuthenticationError(ClientRequestError):
    '''
    Authentication error, access to the resource is forbidden.
    '''
    pass

class ReauthenticateError(ClientRequestError):
    '''
    Authentication error, the client should retry after authenticating again.
    '''
    pass

class ResourceExistsError(ClientRequestError):
    '''
    The resource cannot be created because it exists already.
    '''
    pass

class UnsafeUpdateError(ClientRequestError):
    '''
    The resource cannot be updated safely because it has been modified by
    another client since its state was retrieved, so updating it could
    overwrite these modifications.
    '''
    pass

class NoJobError(ClientRequestError):
    '''
    The request could not return any job.
    '''
    pass

class UnhandledResponseError(ClientError):
    '''
    The response could not be parsed or handled.
    '''
    pass

class ServerError(HTTPError):
    '''
    Server-side exception.
    '''
    pass

def _handle_response_errors(response):
    code = response.status_code
    if code in [200, 201, 204]:
        return
    if code == requests.codes.forbidden:
        raise AuthenticationError(response.text, code)
    if code == requests.codes.unauthorized:
        raise ReauthenticateError(response.text, code)
    if code == requests.codes.precondition_failed:
        raise UnsafeUpdateError(response.text, code)
    if code in range(400, 500):
        raise ClientRequestError(response.text, code)
    if code in range(500, 600):
        raise ServerError(response.text, code)
    raise UnhandledResponseError(response.text, code)




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

import requests
import datetime

_PRE_EXPIRATION_TIME = datetime.timedelta(minutes=1)
_PRE_EXPIRATION_RATIO = 0.95

class JWTTokenAuth(requests.auth.AuthBase):
    def __init__(self, token_string, expires_at):
        self.token_string = token_string
        self.expires_at = expires_at
        cur_time = datetime.datetime.now()
        self._pre_expiration = max(
            self.expires_at - _PRE_EXPIRATION_TIME,
            cur_time + datetime.timedelta(seconds=(self.expires_at - cur_time).total_seconds() * _PRE_EXPIRATION_RATIO),
        )

    def __call__(self, request):
        request.headers['Authorization'] = 'Bearer ' + self.token_string
        return request

    def is_expired(self):
        return datetime.datetime.now() >= self.expires_at

    def expires_soon(self):
        return datetime.datetime.now() >= self._pre_expiration




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

import json
from traceback import format_exc
import warnings

_additional_convert = []

# Only add conversions if the modules exist.
# This way, we do not add unnecessary dependencies
try:
    import numpy as np
    def _np_convert(obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist(), True
        if isinstance(obj, np.generic):
            return np.asscalar(obj), True
        return None, False
    _additional_convert.append(_np_convert)
except ImportError:
    pass

class SchedyJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        for convert in _additional_convert:
            try:
                val, converted = convert(obj)
                if converted:
                    return val
            except Exception:
                warnings.warn(format_exc())
        return super().default(obj)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

from .core import *
from .experiments import *
from .jobs import *




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

import argparse
import schedy
import json
from tabulate import tabulate
import getpass
from requests.compat import urljoin
import os
import stat
import errno
from .compat import json_dumps

DEFAULT_CATEGORY = 'schedy'

def setup_add(subparsers):
    parser = subparsers.add_parser('add', help='Add an experiment.')
    parser.set_defaults(func=cmd_add)
    parser.add_argument('experiment', help='Name for the new experiment.')
    parser.add_argument('-s', '--status', default=schedy.Experiment.RUNNING, choices=(schedy.Experiment.RUNNING, schedy.Experiment.DONE), help='Status of the new experiment.')
    sched_subparsers = parser.add_subparsers(title='Schedulers', dest='scheduler', help='Scheduler type (manual search, random search...)')
    sched_subparsers.required = True
    # Manual scheduling
    sched_subparsers.add_parser('manual', help='Manual search')
    # Random search
    random_parser = sched_subparsers.add_parser('random', help='Random search')
    RANDOM_HP_HELP = (
        'List of the hyperparameters. Each hyperparameter consists in three arguments: name distribution params. '
        '"name" is the name of the hyperparameter. '
        '"distribution" is a distribution among: {}. '
        '"params" is the JSON value for the parameters of the distribution. '
        'Example: learning_rate loguniform \'{{"base": 10, "lowExp": -1, "highExp": 5}}\' num_lays choice \'{{"values": [5, 6, 7, 8]}}\''
    ).format(', '.join(schedy.random._DISTRIBUTION_TYPES.keys()))
    random_parser.add_argument('hyperparameters', nargs='+', help=RANDOM_HP_HELP)
    random_parser.set_defaults(parser=random_parser)

def cmd_add(args):
    db = schedy.SchedyDB(config_path=args.config)
    if args.scheduler == 'manual':
        exp = schedy.ManualSearch(args.experiment, status=args.status)
    elif args.scheduler == 'random':
        if len(args.hyperparameters) % 3 != 0:
            args.parser.error('Invalid hyperparameters (not a list of name/distribution/params).')
        hyperparameters = {}
        for i in range(0, len(args.hyperparameters), 3):
            name = args.hyperparameters[i]
            dist_name = args.hyperparameters[i + 1]
            params_txt = args.hyperparameters[i + 2]
            if name in hyperparameters:
                args.parser.error('Duplicate hyperparameter: {}.'.format(name))
            try:
                dist_type = schedy.random._DISTRIBUTION_TYPES[dist_name]
            except KeyError:
                args.parser.error('Invalid distribution: {}.'.format(dist_name))
            try:
                params = json.loads(params_txt)
                hyperparameters[name] = dist_type._from_args(params)
            except (TypeError, ValueError, KeyError) as e:
                args.parser.error('Invalid distribution parameters for {} ({!r}).'.format(name, e))
        exp = schedy.RandomSearch(args.experiment, status=args.status, distributions=hyperparameters)
    db.add_experiment(exp)

def setup_rm(subparsers):
    parser = subparsers.add_parser('rm', help='Remove an experiment or a job.')
    parser.set_defaults(func=cmd_rm)
    parser.add_argument('experiment', help='Name of the experiment.')
    parser.add_argument('job', nargs='?', help='Name of the job.')
    parser.add_argument('-f', '--force', action='store_true', help='Don\'t ask for confirmation.')

def cmd_rm(args):
    db = schedy.SchedyDB(config_path=args.config)
    exp = db.get_experiment(args.experiment)
    if args.job is None:
        if not args.force:
            print_exp(exp)
            confirmation = input('Are you sure you want to remove {} and all its jobs? [y/N] '.format(exp.name))
            if confirmation.lower() != 'y':
                print('{} was not removed.'.format(exp.name))
                return
        exp.delete()
    else:
        job = exp.get_job(args.job)
        if not args.force:
            print_job(job)
            confirmation = input('Are you sure you want to remove {}? [y/N] '.format(job.job_id))
            if confirmation.lower() != 'y':
                print('{} was not removed.'.format(job.job_id))
                return
        job.delete()

def setup_show(subparsers):
    parser = subparsers.add_parser('show', help='Show an experiment/a job.')
    parser.set_defaults(func=cmd_show)
    parser.add_argument('experiment', help='Name for the new experiment.')
    parser.add_argument('job', nargs='?', help='Name of the job.')

def cmd_show(args):
    db = schedy.SchedyDB(config_path=args.config)
    exp = db.get_experiment(args.experiment)
    if args.job is None:
        print_exp(exp)
    else:
        job = exp.get_job(args.job)
        print_job(job)

def setup_list(subparsers):
    parser = subparsers.add_parser('list', help='List experiments/jobs (by default, lists all experiments).')
    parser.set_defaults(func=cmd_list, parser=parser)
    parser.add_argument('experiment', nargs='?', help='Name of the experiment whose jobs will be listed.')
    parser.add_argument('-t', '--table', action='store_true', help='Long description, as a table.')
    parser.add_argument('-p', '--paragraph', action='store_true', help='Long description, as a series of paragraphs.')
    parser.add_argument('-s', '--sort', action='append', help='Field by which we should sort. You can specify multiple fields using this argument multiple times.')
    parser.add_argument('-d', '--decreasing', action='store_true', help='Sort in reverse order (decreasing values).')
    parser.add_argument('-f', '--field', action='append', help='Specify this option multiple times to select the fields you want to diply (all by default).')

def cmd_list(args):
    db = schedy.SchedyDB(config_path=args.config)
    if args.experiment is None:
        experiments = db.get_experiments()
        table = exp_table(experiments)
    else:
        exp = db.get_experiment(args.experiment)
        jobs = exp.all_jobs()
        table = job_table(jobs)
    if args.sort is not None:
        try:
            table.sort(args.sort, reverse=args.decreasing)
        except KeyError as e:
            args.parser.error(str(e))
    if args.field is not None:
        table.filter_fields(args.field)
    if args.table:
        table.print_table()
    elif args.paragraph:
        table.print_paragraphs()
    else:
        if args.field is None:
            table.filter_categories([DEFAULT_CATEGORY])
        table.print_table('plain', include_headers=False)

def setup_push(subparsers):
    parser = subparsers.add_parser('push', help='Manually add a job to an existing experiment.')
    parser.set_defaults(func=cmd_push)
    parser.add_argument('experiment', help='Name of the experiment for the job.')
    parser.add_argument('-s', '--status', choices=(schedy.Job.QUEUED, schedy.Job.RUNNING, schedy.Job.DONE, schedy.Job.CRASHED, schedy.Job.PRUNED), help='Status of the job.')
    parser.add_argument('-r', '--results', nargs='+', help='Optional results for the job. Each result must be provided as a pair: name value. value must be a valid JSON value. For example: -r accuracy 0.9 loss_history \'[0.9, 0.8, 0.7]\'')
    parser.add_argument('-p', '--hyperparameters', nargs='+', required=True, help='Hyperparameters for the job. Each hyperparameter must be provided as a pair: name value. value must be a valid JSON value. For example: -p learning_rate 0.01 num_layers 3 size_layers \'[512, 1024, 512]\'')
    parser.set_defaults(parser=parser)

def cmd_push(args):
    db = schedy.SchedyDB(config_path=args.config)
    kwargs = dict()
    # Status
    if args.status is not None:
        kwargs['status'] = args.status
    # Results
    if args.results is not None:
        if len(args.results) % 2 != 0:
            args.parser.error('Invalid results (not a list of name/value).')
        results = dict()
        for i in range(0, len(args.results), 2):
            name = args.results[i]
            value_txt = args.results[i + 1]
            if name in results:
                args.parser.error('Duplicate result: {}.'.format(name))
            try:
                results[name] = json.loads(value_txt)
            except (TypeError, ValueError) as e:
                args.parser.error('Invalid value for result {} ({!r}).'.format(name, e))
        kwargs['results'] = results
    # Hyperparameters
    if len(args.hyperparameters) % 2 != 0:
        args.parser.error('Invalid hyperparameters (not a list of name/value).')
    hyperparameters = dict()
    for i in range(0, len(args.hyperparameters), 2):
        name = args.hyperparameters[i]
        value_txt = args.hyperparameters[i + 1]
        if name in hyperparameters:
            args.parser.error('Duplicate hyperparameter: {}.'.format(name))
        try:
            hyperparameters[name] = json.loads(value_txt)
        except (TypeError, ValueError) as e:
            args.parser.error('Invalid value for hyperparameter {} ({!r}).'.format(name, e))
    kwargs['hyperparameters'] = hyperparameters
    exp = db.get_experiment(args.experiment)
    job = exp.add_job(**kwargs)
    print_job(job)

def setup_gen_token(subparsers):
    parser = subparsers.add_parser('gen-token', help='Generate and save a new API token (i.e. configuration file).')
    parser.set_defaults(func=cmd_gen_token)
    parser.add_argument('--root', default='https://api.schedy.io/', help='Schedy API root URL.')
    parser.add_argument('--email', help='User email (prompted if not provided).')
    parser.add_argument('--password', help='User password (prompted if not provided).')

def cmd_gen_token(args):
    config = {
        'root': args.root,
        'token_type': 'password',
    }
    if args.email is None:
        config['email'] = input('Email: ')
    else:
        config['email'] = args.email
    if args.password is None:
        config['token'] = getpass.getpass('Password: ')
    else:
        config['token'] = args.password
    db = schedy.SchedyDB(config_override=config)
    url = urljoin(db.root, 'resettoken/')
    response = db._authenticated_request('POST', url=url)
    schedy.errors._handle_response_errors(response)
    new_content = response.json()
    if args.config is None:
        config_path = schedy.core._default_config_path()
    else:
        config_path = args.config
    config_dir = os.path.dirname(config_path)
    if config_dir:
        try:
            os.makedirs(config_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            pass
    with open(config_path, 'w') as config_file:
        config_file.write(json_dumps(new_content, cls=schedy.encoding.SchedyJSONEncoder))
    try:
        os.chmod(config_path, stat.S_IRUSR | stat.S_IWUSR)
    except OSError:
        print('Token file permissions could not be set.')
    print('Your token has been saved to {}.'.format(config_path))

def main():
    parser = argparse.ArgumentParser(description='Manage your Schedy jobs.')
    parser.add_argument('--config', type=str, help='Schedy configuration file.')
    subparsers = parser.add_subparsers(title='Commands', dest='command')
    subparsers.required = True
    setup_add(subparsers)
    setup_rm(subparsers)
    setup_show(subparsers)
    setup_list(subparsers)
    setup_push(subparsers)
    setup_gen_token(subparsers)
    args = parser.parse_args()
    args.func(args)

class TableData(object):
    def __init__(self):
        self.headers = list()
        self.rows = list()

    def add_row(self, data):
        row = [None] * len(self.headers)
        for key, value in data.items():
            try:
                idx = self.headers.index(key)
            except ValueError:
                idx = len(self.headers)
                self.headers.append(key)
            if idx >= len(row):
                row.extend([None] * (idx - len(row) + 1))
            row[idx] = value
        self.rows.append(row)

    def header_names(self, explicit=False):
        used_names = dict()
        if not explicit:
            expand = [False] * len(self.headers)
            for idx, (category, name) in enumerate(self.headers):
                withoutexp = used_names.setdefault(name, idx)
                withexp = used_names.setdefault(category + '.' + name, idx)
                if withoutexp != idx:
                    expand[idx] = True
                    expand[withoutexp] = True
                if withexp != idx:
                    expand[idx] = True
                    expand[withoutexp] = True
        else:
            expand = [True] * len(self.headers)
        names = []
        for (category, name), should_expand in zip(self.headers, expand):
            if should_expand:
                names.append(category + '.' + name)
            else:
                names.append(name)
        return names

    def _get_fields_indices(self, fields):
        indices = []
        for field in fields:
            try:
                indices.append(self.header_names(explicit=True).index(field))
            except ValueError:
                try:
                    indices.append(self.header_names(explicit=False).index(field))
                except ValueError:
                    raise KeyError('Field "{}" not found or ambiguous.'.format(field))
        return indices

    def _filter_columns(self, indices):
        self.rows = [[row[idx] if idx < len(row) else None for idx in indices] for row in self.rows]
        self.headers = [self.headers[idx] for idx in indices]

    def sort(self, fields, reverse=False):
        indices = self._get_fields_indices(fields)
        def key_func(row):
            key = tuple()
            for idx in indices:
                if idx < len(row):
                    val = row[idx]
                else:
                    val = None
                # Always put None at the end
                if val is None:
                    key = key + (not reverse, None)
                else:
                    key = key + (reverse, val)
            return key
        self.rows = sorted(self.rows, key=key_func, reverse=reverse)

    def filter_categories(self, categories):
        fields_idx = [idx for idx, (cat, _) in enumerate(self.headers) if cat in categories]
        self._filter_columns(fields_idx)

    def filter_fields(self, fields):
        indices = self._get_fields_indices(fields)
        self._filter_columns(indices)

    def print_paragraphs(self):
        fields = self.header_names()
        for i in range(len(self.rows)):
            row = self.rows[i]
            for field, value in zip(fields, row):
                if value is not None:
                    print('{}: {}'.format(field, value))
            if i != len(self.rows) - 1:
                print()

    def print_table(self, fmt='psql', include_headers=True):
        if include_headers:
            print(tabulate(self.rows, self.header_names(), tablefmt=fmt))
        else:
            print(tabulate(self.rows, tablefmt=fmt))

def exp_table(experiments):
    data = TableData()
    for exp in experiments:
        row = {
            (DEFAULT_CATEGORY, 'name'): exp.name,
            (DEFAULT_CATEGORY, 'status'): exp.status,
            (DEFAULT_CATEGORY, 'scheduler'): exp._SCHEDULER_NAME,
        }
        if isinstance(exp, schedy.RandomSearch):
            for name, dist in exp.distributions.items():
                row[('hyperparameter', name)] = '{} ({})'.format(dist._FUNC_NAME, json_dumps(dist._args(), cls=schedy.encoding.SchedyJSONEncoder))
        data.add_row(row)
    return data

def job_table(jobs):
    data = TableData()
    for job in jobs:
        row = {
            (DEFAULT_CATEGORY, 'id'): job.job_id,
            (DEFAULT_CATEGORY, 'status'): job.status,
        }
        for name, value in job.hyperparameters.items():
            row[('hyperparameter', name)] = value
        if job.results is not None:
            for name, value in job.results.items():
                row[('result', name)] = value
        data.add_row(row)
    return data

def print_exp(exp):
    exp_table([exp]).print_paragraphs()

def print_job(job):
    job_table([job]).print_paragraphs()

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

from six import raise_from
from . import errors, encoding
from .compat import json_dumps

def _check_status(status):
    return status in (Job.QUEUED, Job.RUNNING, Job.CRASHED, Job.PRUNED, Job.DONE)

class Job(object):
    #: Status of a queued job. Queued jobs are returned when calling :py:meth:`schedy.Experiment.next_job`.
    QUEUED = 'QUEUED'
    #: Status of a job that is currently running on a worker.
    RUNNING = 'RUNNING'
    #: Status of job that was being processed by a worker, but the worker crashed before completing the job.
    CRASHED = 'CRASHED'
    #: Status of a job that was abandonned because it was not worth working on.
    PRUNED = 'PRUNED'
    #: Status of a completed job.
    DONE = 'DONE'

    def __init__(self, job_id, experiment, hyperparameters, status=QUEUED, results=dict(), etag=None):
        '''
        Represents a job instance belonging to an experiment. You should not
        need to create it by hand. Use :py:meth:`schedy.Experiment.add_job`,
        :py:meth:`schedy.Experiment.get_job`,
        :py:meth:`schedy.Experiment.all_jobs` or
        :py:meth:`schedy.Experiment.next_job` instead.

        Jobs object are context managers, that it to say they can be used with
        a ``with`` statement. They will be put in the RUNNING state at the
        start of the with statement, and in the DONE or CRASHED state at the
        end (depending on whether an uncaught exception is raised within the
        ``with`` block). See :py:meth:`schedy.Job.__enter__` for an example of
        how to use this feature.

        Args:
            job_id (str): Unique id of the job.
            experiment (schedy.Experiment): Experiment containing this job.
            hyperparameters (dict): A dictionnary of hyperparameters values.
            status (str): Job status. See :ref:`job_status`.
            results (dict): A dictionnary of results values.
            etag (str): Value of the entity tag sent by the backend.
        '''
        self.job_id = job_id
        self.experiment = experiment
        self.status = status
        self.hyperparameters = hyperparameters
        self.results = results
        self.etag = etag

    def __str__(self):
        return '{}(id={!r}, experiment={!r}, hyperparameters={!r})'.format(self.__class__.__name__, self.job_id, self.experiment.name, self.hyperparameters)

    def put(self, safe=True):
        '''
        Puts a job in the database, either by creating it or by updating it.

        This function is always called at the end of a ``with`` block.

        Args:
            safe (bool): If true, this operation will make sure not to erase
                any content that would have been put by another Schedy call in
                the meantime. For example, this ensures that no two workers
                overwrite each other's work on this job because they are working
                in parallel.
        '''
        db = self.experiment._db
        url = db._job_url(self.experiment.name, self.job_id)
        map_def = self._to_map_definition()
        data = json_dumps(map_def, cls=encoding.SchedyJSONEncoder)
        headers = dict()
        if safe:
            if self.etag is None:
                headers['If-None-Match'] = '*'
            else:
                headers['If-Match'] = self.etag
        response = db._authenticated_request('PUT', url, data=data, headers=headers)
        errors._handle_response_errors(response)
        etag = response.headers.get('ETag')
        if etag is not None:
            self.etag = etag

    def try_run(self):
        '''
        Try to set the status of the job as ``RUNNING``, or raise an exception
        if another worker tried to do so before this one.
        '''
        self.status = Job.RUNNING
        self.put()

    def delete(self, ensure=True):
        '''
        Deletes this job from the Schedy service.

        Args:
            ensure (bool): If true, an exception will be raised if the job was
                deleted before this call.
        '''
        db = self.experiment._db
        url = db._job_url(self.experiment.name, self.job_id)
        if ensure:
            headers = {'If-Match': '*'}
        else:
            headers = dict()
        response = db._authenticated_request('DELETE', url, headers=headers)
        errors._handle_response_errors(response)

    def __enter__(self):
        '''
        Context manager ``__enter__`` method. Will try to set the job as
        ``CRASHED`` if the job has not been modified by another worker
        concurrently.

        Example:

        >>> db = schedy.SchedyDB()
        >>> exp = db.get_experiment('Test')
        >>> with exp.next_job() as job:
        >>>     my_train_function(job)

        If ``my_train_function`` raises an exception, the job will be marked as
        ``CRASHED``. Otherwise it will be marked as ``DONE``. (See
        py:meth:`Job.__exit__`.)

        Note that since :py:meth:`schedy.Experiment.next_job` will always return
        a ``RUNNING`` job, this method will never raise
        :py:exc:`schedy.errors.UnsafeUpdateError` in this case.
        '''
        if self.status != Job.RUNNING:
            self.try_run()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''
        Context manager ``__exit__`` method. Will try to set the job status as
        ``CRASHED`` if an exception was raised in the ``with`` block.
        Otherwise, it will try to set the job status as ``DONE``. It will also
        push all the updates that were made locally to the Schedy service (by
        calling :py:meth:`Job.put` for you).
        '''
        if exc_type is not None:
            self.status = Job.CRASHED
        else:
            self.status = Job.DONE
        self.put()

    @classmethod
    def _from_map_definition(cls, experiment, map_def, etag=None):
        try:
            job_id = str(map_def['id'])
            experiment_name = str(map_def['experiment'])
            status = str(map_def['status'])
            hyperparameters = map_def.get('hyperparameters')
            if hyperparameters is not None:
                hyperparameters = dict(hyperparameters)
            else:
                hyperparameters = dict()
            results = map_def.get('results')
            if results is not None:
                results = dict(results)
            else:
                results = dict()
        except (KeyError, ValueError) as e:
            raise_from(ValueError('Invalid job map definition.'), e)
        if experiment_name != experiment.name:
            raise ValueError('Inconsistent experiment name for job: expected {}, found {}.'.format(experiment.name, experiment_name))
        if not _check_status(status):
            raise ValueError('Invalid or unknown status value: {}.'.format(status))
        return cls(
                job_id=job_id,
                experiment=experiment,
                status=status,
                hyperparameters=hyperparameters,
                results=results,
                etag=etag)

    def _to_map_definition(self):
        map_def = {
                'status': str(self.status),
            }
        if len(self.hyperparameters) > 0:
            map_def['hyperparameters'] = self.hyperparameters
        if self.results is not None and len(self.results) > 0:
            map_def['results'] = self.results
        return map_def

def _make_job(experiment, data, etag=None):
    try:
        job_data = dict(data)
    except ValueError as e:
        raise_from(errors.UnhandledResponseError('Excepting the description of a job as a dict, received type {}.'.format(type(data)), None), e)
    try:
        job = Job._from_map_definition(experiment, job_data, etag)
    except ValueError as e:
        raise_from(errors.UnhandledResponseError('Response contains an invalid job.', None), e)
    return job

def _job_from_response(experiment, response):
    try:
        content = response.json()
    except ValueError as e:
        raise_from(errors.UnhandledResponseError('Response contains invalid JSON:\n' + response.text, None), e)
    return _make_job(experiment, content, response.headers.get('ETag'))




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

class LogUniform(object):
    _FUNC_NAME = 'loguniform'

    def __init__(self, low, high):
        '''
        LogUniform distribution. Values are sampled betweend ``low`` and ``high``,
        such that ``log(value)`` is uniformly distributed between ``log(low)`` and
        ``log(high)``.

        Args:
            low (float): Minimal value (inclusive).
            high (float): Maximum value (exclusive).
        '''
        self.low = low
        self.high = high

    def _args(self):
        return {
            'low': float(self.low),
            'high': float(self.high),
        }

    @classmethod
    def _from_args(cls, args):
        low = float(args['low'])
        high = float(args['high'])
        return cls(low, high)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.low == other.low and \
            self.high == other.high

class Uniform(object):
    _FUNC_NAME = 'uniform'

    def __init__(self, low, high):
        '''
        Uniform distribution. Values will be uniformly distributed in the interval [``low``, ``high``).

        Args:
            low (float): Minimal value (inclusive).
            high (float): Maximum value (exclusive).
        '''
        self.low = low
        self.high = high

    def _args(self):
        return {
            'low': float(self.low),
            'high': float(self.high),
        }

    @classmethod
    def _from_args(cls, args):
        low = float(args['low'])
        high = float(args['high'])
        return cls(low, high)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.low == other.low and \
            self.high == other.high

class Choice(object):
    _FUNC_NAME = 'choice'

    def __init__(self, values, weights=None):
        '''
        Choice distribution. Values will be picked randomly in a set of values. You can optionally provide
        weights for these values, to make some of them more likely to be suggested by
        Schedy than others.

        Args:
            values (list): Possible values that can be picked. They can be
                numbers, strings, booleans, strings, lists or dictionaries, and
                you can mix those.
            weights (list): Weight associated with each value. If provided, the
                length of ``weights`` must be the same as that of ``values``.
        '''
        self.values = values
        self.weights = weights

    def _args(self):
        args = {
            'values': list(self.values),
        }
        if self.weights is not None:
            if len(self.weights) != len(self.values):
                raise ValueError('There must be as many weights as there are values.')
            args['weights'] = [float(weight) for weight in self.weights]
        return args

    @classmethod
    def _from_args(cls, args):
        values = list(args['values'])
        weights = None
        weights_val = args.get('weights')
        if weights_val != None:
            weights = [float(w) for w in weights_val]
        return cls(values, weights)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.values == other.values and \
            self.weights == other.weights

class Normal(object):
    _FUNC_NAME = 'normal'

    def __init__(self, mean, std):
        '''
        Normal distribution.

        Args:
            mean (float): Desired mean of the distribution.
            std (float): Desired standard deviation of the distribution.
        '''
        self.mean = mean
        self.std = std

    def _args(self):
        return {
            'mean': float(self.mean),
            'std': float(self.std),
        }

    @classmethod
    def _from_args(cls, args):
        mean = float(args['mean'])
        std = float(args['std'])
        return cls(mean, std)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.mean == other.mean and \
            self.std == other.std

class Constant(object):
    _FUNC_NAME = 'const'

    def __init__(self, value):
        '''
        "Constant" distribution. Will always yield the same value.

        Args:
            value: The value of the samples that will be returned by this
                distribution. Can be a number, string, boolean, string, list or
                dictionary.
        '''
        self.value = value

    def _args(self):
        return self.value

    @classmethod
    def _from_args(cls, args):
        val = args
        return cls(val)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.value == other.value

_DISTRIBUTION_TYPES = {cls._FUNC_NAME: cls for cls in (LogUniform, Uniform, Choice, Normal, Constant)}



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *
from six import raise_from

import warnings

from . import errors

_EXPECTED_PAGE_KEYS = {'items', 'next'}

class PageObjectsIterator(object):
    def __init__(self, reqfunc, obj_creation_func):
        self._reqfunc = reqfunc
        self._create_obj = obj_creation_func
        self._next_token = None
        self._items = []
        self._get_page()

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._items) == 0:
            if self._next_token is None:
                raise StopIteration
            self._get_page(self._next_token)
            if len(self._items) == 0:
                raise StopIteration
        cur_item = self._create_obj(self._items[0])
        self._items = self._items[1:]
        return cur_item

    def _get_page(self, start_token=None):
        if start_token is None:
            params = dict()
        else:
            params = {'start': start_token}
        response = self._reqfunc(params)
        errors._handle_response_errors(response)
        try:
            result = dict(response.json())
        except ValueError as e:
            raise_from(errors.UnhandledResponseError('Expected page as a dict.'), e)
        if result.keys() > _EXPECTED_PAGE_KEYS:
            warnings.warn('Unexpected page keys: {}.'.format(result.keys() - _EXPECTED_PAGE_KEYS))
        try:
            self._items = list(result['items'])
            next_token = result.get('next')
            if next_token is not None:
                self._next_token = str(next_token)
            else:
                self._next_token = None
        except (ValueError, KeyError) as e:
            raise_from(errors.UnhandledResponseError('Invalid page received.'), e)




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *
from six import raise_from

import requests
from requests.compat import urljoin
import functools
import logging

from . import errors, encoding
from .random import _DISTRIBUTION_TYPES
from .pbt import _EXPLOIT_STRATEGIES, _EXPLORE_STRATEGIES
from .jobs import Job, _make_job, _job_from_response
from .pagination import PageObjectsIterator
from .compat import json_dumps

logger = logging.getLogger(__name__)

def _check_status(status):
    return status in (Experiment.RUNNING, Experiment.DONE)

class Experiment(object):
    #: Status of a running experiment.
    RUNNING = 'RUNNING'
    #: Status of a completed (or paused) experiment.
    DONE = 'DONE'

    def __init__(self, name, status=RUNNING):
        '''
        Base-class for all experiments.

        Args:
            name (str): Name of the experiment. An experiment is uniquely
                identified by its name.
            status (str): Status of the experiment. See :ref:`experiment_status`.
        '''
        self.name = name
        self.status = status
        self._db = None

    def add_job(self, **kwargs):
        '''
        Adds a new job to this experiment.

        Args:
            hyperparameters (dict): A dictionnary of hyperparameters values.
            status (str): Job status. See :ref:`job_status`. Default: QUEUED.
            results (dict): A dictionnary of result values. Default: No results (empty dictionary).

        Returns:
            schedy.Job: The instance of the new job.
        '''
        partial_job = Job(
                job_id=None,
                experiment=None,
                **kwargs)
        assert self._db is not None, 'Experiment was not added to a database'
        url = self._jobs_url()
        map_def = partial_job._to_map_definition()
        data = json_dumps(map_def, cls=encoding.SchedyJSONEncoder)
        response = self._db._authenticated_request('POST', url, data=data)
        errors._handle_response_errors(response)
        return _job_from_response(self, response)

    def next_job(self):
        '''
        Returns a new job to be worked on. This job will be set in the
        ``RUNNING`` state. This function handles everything so that two
        workers never start working on the same job.

        Returns:
            schedy.Job: The instance of the requested job.
        '''
        assert self._db is not None, 'Experiment was not added to a database'
        url = urljoin(self._db._experiment_url(self.name), 'nextjob/')
        job = None
        # Try obtaining a job and running it until we manage to get hold of a
        # job we can indeed run (concurrent trials to run a job can cause us to
        # fail, so try and try again)
        while job is None:
            response = self._db._authenticated_request('GET', url)
            if response.status_code == requests.codes.no_content:
                raise errors.NoJobError('No job left for experiment {}.'.format(self.name), None)
            errors._handle_response_errors(response)
            job = _job_from_response(self, response)
            try:
                job.try_run()
            except errors.UnsafeUpdateError:
                job = None
                logger.debug('Two workers tried to start working on the same job, retrying.', exc_info=True)
        return job

    def all_jobs(self):
        '''
        Retrieves all the jobs belonging to this experiment.

        Returns:
            iterator of :py:class:`schedy.Job`: An iterator over all the jobs of this experiment.
        '''
        assert self._db is not None, 'Experiment was not added to a database'
        url = self._jobs_url()
        return PageObjectsIterator(
            reqfunc=functools.partial(self._db._authenticated_request, 'GET', url),
            obj_creation_func=functools.partial(_make_job, self),
        )

    def get_job(self, job_id):
        '''
        Retrieves a job by id.

        Args:
            job_id (str): Id of the job to retrieve.

        Returns:
            schedy.Job: Instance of the requested job.
        '''
        assert self._db is not None, 'Experiment was not added to a database'
        url = self._db._job_url(self.name, job_id)
        response = self._db._authenticated_request('GET', url)
        errors._handle_response_errors(response)
        job = _job_from_response(self, response)
        return job

    def __str__(self):
        try:
            return '{}(name={!r}, params={})'.format(self.__class__.__name__, self.name, self._get_params())
        except NotImplementedError:
            return '{}(name={!r})'.format(self.__class__.__name__, self.name)

    def push_updates(self):
        '''
        Push all the updates made to this experiment to the service.
        '''
        assert self._db is not None, 'Experiment was not added to a database'
        url = self._db._experiment_url(self.name)
        content = self._to_map_definition()
        data = json_dumps(content, cls=encoding.SchedyJSONEncoder)
        response = self._db._authenticated_request('PUT', url, data=data)
        errors._handle_response_errors(response)

    def delete(self, ensure=True):
        '''
        Deletes this experiment.

        Args:
            ensure (bool): If true, an exception will be raised if the experiment was
                deleted before this call.
        '''
        assert self._db is not None, 'Experiment was not added to a database'
        url = self._db._experiment_url(self.name)
        if ensure:
            headers = {'If-Match': '*'}
        else:
            headers = dict()
        response = self._db._authenticated_request('DELETE', url, headers=headers)
        errors._handle_response_errors(response)

    @classmethod
    def _create_from_params(cls, name, status, params):
        raise NotImplementedError()

    def _get_params(self):
        raise NotImplementedError()

    def _to_map_definition(self):
        try:
            scheduler = self._SCHEDULER_NAME
        except AttributeError as e:
            raise_from(AttributeError('Experiment implementations should define a _SCHEDULER_NAME attribute'), e)
        return {
                'status': self.status,
                'scheduler': {scheduler: self._get_params()},
            }

    @staticmethod
    def _from_map_definition(schedulers, map_def):
        try:
            name = str(map_def['name'])
            status = str(map_def['status'])
            scheduler_def_map = dict(map_def['scheduler'])
            if len(scheduler_def_map) != 1:
                raise ValueError('Invalid scheduler definition: {}.'.format(scheduler_def_map))
            sched_def_key, sched_def_val = next(iter(scheduler_def_map.items()))
            scheduler = str(sched_def_key)
            params = sched_def_val
        except (ValueError, KeyError) as e:
            raise_from(ValueError('Invalid map definition for experiment.'), e)
        if not _check_status(status):
            raise ValueError('Invalid or unknown status value: {}.'.format(status))
        try:
            exp_type = schedulers[scheduler]
        except KeyError as e:
            raise ValueError('Invalid or unregistered scheduler name: {}.'.format(scheduler))
        return exp_type._create_from_params(
                name=name,
                status=status,
                params=params)

    def _jobs_url(self):
        return urljoin(self._db._experiment_url(self.name), 'jobs/')

class ManualSearch(Experiment):
    '''
    Represents a manual search, that is to say an experiment for which the only
    jobs returned by :py:meth:`schedy.Experiment.next_job` are jobs that were
    queued beforehand (by using :py:meth:`schedy.Experiment.add_job` for
    example).
    '''
    _SCHEDULER_NAME = 'Manual'

    @classmethod
    def _create_from_params(cls, name, status, params):
        if params is not None:
            raise ValueError('Expected no parameters for manual search, found {}.'.format(type(params)))
        return cls(name=name, status=status)

    def _get_params(self):
        return None

class RandomSearch(Experiment):
    _SCHEDULER_NAME = 'RandomSearch'

    def __init__(self, name, distributions, status=Experiment.RUNNING):
        '''
        Represents a random search, that is to say en experiment that returns
        jobs with random hyperparameters when no job was queued manually using
        :py:meth:`schedy.Experiment.add_job`.

        If you create a job manually for this experiment, it must have only and
        all the hyperparameters specified in the ``distributions`` parameter.

        Args:
            name (str): Name of the experiment. An experiment is uniquely
                identified by its name.
            distributions (dict): A dictionary of distributions (see
                :py:mod:`schedy.random`), whose keys are the names of the
                hyperparameters.
            status (str): Status of the experiment. See :ref:`experiment_status`.
        '''
        super().__init__(name, status)
        self.distributions = distributions

    @classmethod
    def _create_from_params(cls, name, status, params):
        try:
            items = params.items()
        except AttributeError as e:
            raise ValueError('Expected parameters as a dict, found {}.'.format(type(params)))
        distributions = dict()
        for key, dist_def in items:
            try:
                dist_name_raw, dist_args = next(iter(dist_def.items()))
                dist_name = str(dist_name_raw)
            except (KeyError, TypeError) as e:
                raise_from(ValueError('Invalid distribution definition.'), e)
            try:
                dist_type = _DISTRIBUTION_TYPES[dist_name]
            except KeyError as e:
                raise ValueError('Invalid or unknown distribution type: {}.'.format(dist_name))
            distributions[key] = dist_type._from_args(dist_args)
        return cls(name=name, distributions=distributions, status=status)

    def _get_params(self):
        return {key: {dist._FUNC_NAME: dist._args()} for key, dist in self.distributions.items()}

class PopulationBasedTraining(Experiment):
    _SCHEDULER_NAME = 'PBT'

    def __init__(self, name, objective, result_name, exploit, explore=dict(), initial_distributions=dict(), population_size=None, status=Experiment.RUNNING, max_generations=None):
        '''
        Implements Population Based Training (see `paper
        <https://arxiv.org/pdf/1711.09846.pdf>`_).

        You have two ways to specify the initial jobs for Population Based
        training. You can create them manually using
        :py:meth:`schedy.Experiment.add_job`, or you can specify the
        ``initial_distributions`` and ``population_size`` parameters.

        If you create a job manually for this experiment, it must have at least
        the hyperparameters specified in the ``explore`` parameter.

        Args:
            name (str): Name of the experiment. An experiment is uniquely
                identified by its name.
            objective (str): The objective of the training, either
                :py:attr:`schedy.pbt.MINIMIZE` (to
                minimize a result) or
                :py:attr:`schedy.pbt.MAXIMIZE` (to
                maximize a result).
            result_name (str): The name of the result to optimize. This result
                must be present in the results of all
                :py:attr:`~schedy.Experiment.RUNNING` jobs of this experiment.
            exploit (schedy.pbt.ExploitStrategy): Strategy to use to
                exploit the results (i.e. to focus on the most promising jobs).
            explore (dict): Strategy to use to explore new hyperparameter values.
                The keys of the dictionary are the name of the
                hyperparameters (str), and the values are the strategy
                associated with the hyperparameter
                (:py:class:`schedy.pbt.ExploreStrategy`). Values for the
                omitted hyperparameters will not be explored. This parameter is
                optional: if you do not specify any explore strategy, only
                exploitation will be used.
            initial_distributions (dict): The initial distributions for the
                hyperparameters, as dictionary of distributions (see
                :py:mod:`schedy.random`) whose keys are the names of the
                hyperparameters. This parameter optional, you can also create
                the initial jobs manually. If you use this parameter, make sure
                to use ``population_size`` as well.
            population_size (int): Number of initial jobs to create, before
                starting to exploit/explore (i.e. size of the population). It
                does **not** have to be the number of jobs you can process in
                parallel. The original paper used values between 10 and 80.
            status (str): Status of the experiment. See :ref:`experiment_status`.
            max_generations (int): Maximum number of generations to run before
                marking the experiment the experiments as done
                (:py:ref:`experiment_status`). When the maximum number of
                generations is reached, subsequent calls to
                :py:meth:`schedy.Experiment.next_job` will raise
                :py:class:`schedy.errors.NoJobError`, to indicate that the job
                queue is empty.
        '''
        super().__init__(name, status)
        self.objective = objective
        self.result_name = result_name
        self.exploit = exploit
        self.explore = explore
        self.initial_distributions = initial_distributions
        self.population_size = population_size
        self.max_generations = max_generations

    @classmethod
    def _create_from_params(cls, name, status, params):
        kwargs = {
            'name': name,
            'status': status,
            'objective': params['objective'],
            'result_name': params['qualityResultName'],
        }
        exploit_type, exploit_params = next(iter(params['exploit'].items()))
        kwargs['exploit'] = _EXPLOIT_STRATEGIES[exploit_type]._from_params(exploit_params)
        population_size = params.get('numParallel')
        if population_size is not None:
            kwargs['population_size'] = population_size
        initial_distributions = params.get('init')
        if initial_distributions is not None:
            init_param = dict()
            for hp, dist_map in initial_distributions.items():
                dist_name, dist_params = next(iter(dist_map.items()))
                init_param[hp] = _DISTRIBUTION_TYPES[dist_name]._from_args(dist_params)
            kwargs['initial_distributions'] = init_param
        explore = params.get('explore')
        if explore is not None:
            explore_map = dict()
            for hp, strat_map in explore.items():
                strat_name, strat_params = next(iter(strat_map.items()))
                explore_map[hp] = _EXPLORE_STRATEGIES[strat_name]._from_params(strat_params)
            kwargs['explore'] = explore_map
        max_generations = params.get('max_generations')
        if max_generations is not None:
            kwargs['max_generations'] = max_generations
        return cls(**kwargs)

    def _get_params(self):
        params = {
            'objective': self.objective,
            'qualityResultName': self.result_name,
        }
        if self.population_size:
            params['numParallel'] = self.population_size
        if self.initial_distributions:
            params['init'] = {
                name: {dist._FUNC_NAME: dist._args()} for name, dist in self.initial_distributions.items()
            }
        params['exploit'] = {self.exploit._EXPLOIT_STRATEGY_NAME: self.exploit._get_params()}
        if self.explore:
            params['explore'] = {
                name: {strat._EXPLORE_STRATEGY_NAME: strat._get_params()} for name, strat in self.explore.items()
            }
        if self.max_generations:
            params['maxGenerations'] = self.max_generations
        return params

def _make_experiment(db, data):
    try:
        exp_data = dict(data)
    except ValueError as e:
        raise_from(errors.UnhandledResponseError('Expected experiment data as a dict, received {}.'.format(type(data)), None), e)
    try:
        exp = Experiment._from_map_definition(db._schedulers, exp_data)
    except ValueError as e:
        raise_from(errors.UnhandledResponseError('Response contains an invalid experiment', None), e)
    exp._db = db 
    return exp




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

#: Minimize the objective
MINIMIZE = 'min'
#: Maximize the objective
MAXIMIZE = 'max'

class Truncate(object):
    _EXPLOIT_STRATEGY_NAME = 'truncate'

    def __init__(self, proportion=0.2):
        '''
        Truncate exploit strategy: if the selected candidate job is in the
        worst n%, use a candidate job in the top n% instead.

        Args:
            proportion (float): Proportion of jobs that are considered to be
                "best" jobs, and "worst" jobs. For example, if ``proportion =
                0.2``, if the selected candidate job is in the bottom 20%, it
                will be replaced by a job in the top 20%. Must satisfy ``0 <
                proportion <= 0.5``.
        '''
        self.proportion = proportion

    def _get_params(self):
        return self.proportion

    @classmethod
    def _from_params(cls, params):
        proportion = float(params)
        return cls(proportion)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.proportion == other.proportion

class Perturb(object):
    _EXPLORE_STRATEGY_NAME = 'perturb'

    def __init__(self, min_factor=0.8, max_factor=1.2):
        '''
        Perturb explore strategy: multiply the designated hyperparameter by a
        random factor, sampled from a uniform distribution.

        Args:
            min_factor (float): Minimum value for the factor (inclusive).
            max_factor (float): Maximum value for the factor (exclusive).
        '''
        self.min_factor = min_factor
        self.max_factor = max_factor

    def _get_params(self):
        return {
            'minFactor': float(self.min_factor),
            'maxFactor': float(self.max_factor),
        }

    @classmethod
    def _from_params(cls, params):
        min_factor = float(params['minFactor'])
        max_factor = float(params['maxFactor'])
        return cls(min_factor, max_factor)

    def __eq__(self, other):
        return type(self) == type(other) and \
            self.min_factor == other.min_factor and \
            self.max_factor == other.max_factor

_EXPLOIT_STRATEGIES = {strat._EXPLOIT_STRATEGY_NAME: strat for strat in [
    Truncate
]}

_EXPLORE_STRATEGIES = {strat._EXPLORE_STRATEGY_NAME: strat for strat in [
    Perturb
]}




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *
from six import raise_from

from .experiments import Experiment, RandomSearch, ManualSearch, PopulationBasedTraining, _make_experiment
from .jwt import JWTTokenAuth
from .pagination import PageObjectsIterator
from . import errors, encoding
from .compat import json_dumps

import functools
import json
import requests
import os.path
import datetime
from requests.compat import urljoin, quote as urlquote
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import logging

logger = logging.getLogger(__name__)

class SchedyRetry(Retry):
    BACKOFF_MAX = 8 * 60

    def increment(self, method=None, url=None, response=None, error=None, *args, **kwargs):
        logger.warn('Error while querying Schedy service, retrying.')
        if response is not None:
            logger.warn('Server message: {!s}'.format(response.data))
        return super().increment(
            method=method,
            url=url,
            response=response,
            error=error,
            *args,
            **kwargs)

#: Number of retries if the authentication fails.
NUM_AUTH_RETRIES = 2

def _default_config_path():
    return os.path.join(os.path.expanduser('~'), '.schedy', 'client.json')

class SchedyDB(object):
    def __init__(self, config_path=None, config_override=None):
        '''
        SchedyDB is the central component of Schedy. It represents your
        connection the the Schedy service.

        Args:
            config_path (str or file-object): Path to the client configuration file. This file
                contains your credentials (email, API token). By default,
                ~/.schedy/client.json is used. See :ref:`setup` for
                instructions about how to use this file.
            config_override (dict): Content of the configuration. You can use this to
                if you do not want to use a configuration file.
        '''
        self._load_config(config_path, config_override)
        # Add the trailing slash if it's not there
        if len(self.root) == 0 or self.root[-1] != '/':
            self.root = self.root + '/'
        self._schedulers = dict()
        self._register_default_schedulers()
        self._jwt_token = None
        self._jwt_expiration = datetime.datetime(year=1970, month=1, day=1)
        self._session = None

    def _authenticate(self):
        '''
        Renew authentication. You do not usually need to call this function, as
        it will always be called automatically when needed.
        '''
        logger.debug('Renewing authentication')
        if self.token_type == 'password':
            url = urljoin(self.root, 'passauth/')
        else:
            url = urljoin(self.root, 'token/')
        response = self._perform_request('POST', url, json={'email': self.email, 'token': self.api_token})
        errors._handle_response_errors(response)
        try:
            token_data = response.json()
        except ValueError as e:
            raise_from(errors.ServerError('Response contains invalid JSON:\n' + response.text, None), e)
        try:
            jwt_token = token_data['token']
            expires_at = datetime.datetime.fromtimestamp(token_data['expiresAt'])
        except (KeyError, OverflowError, OSError) as e:
            raise_from(errors.ServerError('Response contains invalid token data.', None), e)
        self._jwt_token = JWTTokenAuth(jwt_token, expires_at)
        logger.debug('A new token was obtained.')

    def add_experiment(self, exp):
        '''
        Adds an experiment to the Schedy service. Use this function to create
        new experiments.

        Args:
            exp (schedy.Experiment): The experiment to add.

        Example:
            >>> db = schedy.SchedyDB()
            >>> exp = schedy.ManualSearch('TestExperiment')
            >>> db.add_experiment(exp)
        '''
        url = self._experiment_url(exp.name)
        content = exp._to_map_definition()
        data = json_dumps(content, cls=encoding.SchedyJSONEncoder)
        response = self._authenticated_request('PUT', url, data=data, headers={'If-None-Match': '*'})
        # Handle code 412: Precondition failed
        if response.status_code == requests.codes.precondition_failed:
            raise errors.ResourceExistsError(response.text, response.status_code)
        else:
            errors._handle_response_errors(response)
        exp._db = self

    def get_experiment(self, name):
        '''
        Retrieves an experiment from the Schedy service by name.

        Args:
            name (str): Name of the experiment.

        Returns:
            schedy.Experiment: An experiment of the appropriate type.
        
        Example:
            >>> db = schedy.SchedyDB()
            >>> exp = db.get_experiment('TestExperiment')
            >>> print(type(exp))
            <class 'schedy.experiments.ManualSearch'>
        '''
        url = self._experiment_url(name)
        response = self._authenticated_request('GET', url)
        errors._handle_response_errors(response)
        try:
            content = dict(response.json())
        except ValueError as e:
            raise_from(errors.ServerError('Response contains invalid JSON dict:\n' + response.text, None), e)
        try:
            exp = Experiment._from_map_definition(self._schedulers, content)
        except ValueError as e:
            raise_from(errors.ServerError('Response contains an invalid experiment', None), e)
        exp._db = self
        return exp

    def get_experiments(self):
        '''
        Retrieves all the experiments from the Schedy service.

        Returns:
            iterator of :py:class:`schedy.Experiment`: Iterator over all the experiments.
        '''
        return PageObjectsIterator(
            reqfunc=functools.partial(self._authenticated_request, 'GET', self._all_experiments_url()),
            obj_creation_func=functools.partial(_make_experiment, self),
        )

    def _register_scheduler(self, experiment_type):
        '''
        Registers a new type of experiment. You should never have to use this
        function yourself.

        Args:
            experiment_type (class): Type of the experiment, it must have an
                attribute called _SCHEDULER_NAME.
        '''
        self._schedulers[experiment_type._SCHEDULER_NAME] = experiment_type

    def _register_default_schedulers(self):
        self._register_scheduler(RandomSearch)
        self._register_scheduler(ManualSearch)
        self._register_scheduler(PopulationBasedTraining)

    def _all_experiments_url(self):
        return urljoin(self.root, 'experiments/')

    def _experiment_url(self, name):
        return urljoin(self._all_experiments_url(), '{}/'.format(urlquote(name, safe='')))

    def _job_url(self, experiment, job):
        return urljoin(self.root, 'experiments/{}/jobs/{}/'.format(urlquote(experiment, safe=''), urlquote(job, safe='')))

    def _load_config(self, config_path, config):
        if config is None:
            if config_path is None:
                config_path = _default_config_path()
            if hasattr(config_path, 'read'):
                config = json.loads(config_path.read())
            else:
                with open(config_path) as f:
                    config = json.load(f)
        self.root = config['root']
        self.email = config['email']
        self.token_type = config.get('token_type', 'api_token')
        allowed_token_types = ['api_token', 'password']
        if self.token_type not in allowed_token_types:
            raise ValueError('Configuration value token_type must be one of {}.'.format(', '.join(allowed_token_types)))
        self.api_token = config['token']

    def _authenticated_request(self, *args, **kwargs):
        response = None
        for _ in range(NUM_AUTH_RETRIES):
            if self._jwt_token is None or self._jwt_token.expires_soon():
                self._authenticate()
            response = self._perform_request(*args, auth=self._jwt_token, **kwargs)
            if response.status_code != requests.codes.unauthorized:
                break
        return response

    def _make_session(self):
        self._session = requests.Session()
        retry_mgr = SchedyRetry(
                total=10,
                read=10,
                connect=10,
                backoff_factor=0.4,
                status_forcelist=frozenset((requests.codes.server_error, requests.codes.unavailable)),
                # Careful: POST and PATCH are in the whitelist. This means that
                # the server should not be in an incomplete state or POSTING
                # and PATCHING twice could do weird things. We do this because
                # we do not want Schedy to crash in the face of the user when
                # there's a connection or benign error.
                method_whitelist=frozenset(('HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE', 'POST', 'PATCH')),
            )
        adapter = HTTPAdapter(max_retries=retry_mgr)
        self._session.mount('http://', adapter)
        self._session.mount('https://', adapter)

    def _perform_request(self, *args, **kwargs):
        if self._session is None:
            self._make_session()
        if 'data' in kwargs:
            logger.debug('Sent headers: %s', kwargs.get('headers'))
            logger.debug('Sent data: %s', kwargs['data'])
        req = self._session.request(*args, **kwargs)
        logger.debug('Received headers: %s', req.headers)
        logger.debug('Received data: %s', req.text)
        return req




#!/usr/bin/env python

from setuptools import setup

setup(
    name='Schedy',
    version='0.1.0b5',
    description='Python client for Schedy',
    long_description=open('README.rst').read(),
    author='Incalia',
    author_email='nathan@incalia.fr',
    url='https://schedy.io/',
    install_requires=[
        'requests>=2.18.4',
        'tabulate>=0.8.2',
        'six>=1.11.0',
        'future>=0.16.0',
    ],
    packages=['schedy'],
    entry_points={
        'console_scripts': [
            'schedy = schedy.cmd:main',
        ],
    },
    license='MIT',
    # TODO: Add license to classifiers
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={
        'Homepage': 'https://schedy.io/',
        'Source': 'https://github.com/incalia/schedy-client/',
        'Documentation': 'https://schedy.readthedocs.io/en/latest/',
    },
)




# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import *

import json

def json_dumps(*args, **kwargs):
    return str(json.dumps(*args, **kwargs))

